import logging
from typing import Any

try:
    from logging import NullHandler
except ImportError:

    class NullHandler(logging.Handler):  # type: ignore[no-redef]
        def emit(self, record: Any) -> Any:
            pass


log = logging.getLogger(__name__)
log.addHandler(NullHandler())
log.propagate = False  # Disabling the transmission of logs higher up the chain


def switch_logger(
    state: bool,
    handler: logging.Handler | None = None,
    formatter: logging.Formatter | None = None,
    level: str = 'DEBUG',
) -> None:
    if state:
        console_handler = handler or logging.StreamHandler()
        console_formatter = formatter or logging.Formatter(
            '%(filename)s:%(lineno)d | def %(funcName)s | %(message)s'
        )

        console_handler.setFormatter(console_formatter)
        log.addHandler(console_handler)
        log.setLevel(getattr(logging, level))
    else:
        log.addHandler(NullHandler)  # type: ignore[arg-type]



import asyncio
import uuid
from collections.abc import Callable, Iterable
from dataclasses import dataclass, field


# TODO(Ars): Отделить поля паспорта от полей справочника
@dataclass
class PassportState:
    pass


@dataclass
class DictState:
    dict_filename: str
    model: str  # dict model
    oid: str  # passport oid
    total_dicts: int
    passport_semaphore: asyncio.Semaphore
    passport_name: str = ''
    passport_pk: str = field(default_factory=lambda: str(uuid.uuid4()))
    dict_model_data: dict[str, dict] = field(default_factory=dict)
    filter: Callable[[dict], bool] | None = None
    include: Iterable[str] | None = None
    exclude: Iterable[str] | None = None
    version_changed: bool = False
    create_sql: bool = False
    dict_changed: bool = False
    forced_update: bool = False

    def __post_init__(self):
        if self.filter is not None and not callable(self.filter):
            raise TypeError('Объект filter должен быть вызываемым - Callable')
        if self.include and self.exclude:
            raise ValueError('Одновременное использование параметров include и exclude запрещено.')

    def __repr__(self) -> str:
        return f'<DictState(dict_filename={self.dict_filename})>'



import abc
from datetime import datetime, timezone
from typing import Any

from .dtos import DictState
from .exceptions import NsiPkNotFoundError
from .file_handlers import FileHandler
from .logger import log


class Exampler(abc.ABC):
    """"""

    @abc.abstractmethod
    def show_passport_model(
        self, passports_modelname: str, dict_internal_pk_field: str, passport_changed: bool
    ) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def show_dict_model(self, dict_state: DictState) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def upd_passport_model_data_with_passport_fields(self, local_passports: list[dict]) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def upd_dict_model_data_with_passport_fields(
        self, dict_state: DictState, remote_passport: dict
    ) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def upd_dict_model_data_with_dict_fields(
        self, dict_state: DictState, dict_data: dict[str, Any]
    ) -> None:
        raise NotImplementedError


# TODO(Ars): Необходимо решить проблему при которой
# при обновлении части справочников ими перезатирается файл с примерами всех справочников.
# Как вариант можно разбивать файл примеров по \n\n находить нужный справочник в списке
# и обновлять только этот элемент в списке, затем перезаписывать файл с 0 всеми справочниками из списка.
class ModelExampler(Exampler):
    """"""

    py_type_to_field_cls_name = {
        type(None): '!!!!!!!!!!!!!!!!!!!! UNKNOWN TYPE !!!!!!!!!!!!!!!!!!!!',
        # NOTE(Ars): Можно добавить TextField для особо больших max_length
        str: 'CharField',
        # NOTE(Ars): В принципе если считать макс значение integer
        # то можно будет выбирать SmallIntegerField или IntegerField или BigIntegerField
        int: 'IntegerField',
        bool: 'BooleanField',
        dict: 'JSONField',
        list: 'JSONField',
    }
    DICT_PK_NAMES: list[str]

    def __init__(
        self,
        file: FileHandler,
        model_examples: str | None,
        model_examples_params: str | None,
        models_filename: str = 'models_example.py',
        parent_dict_cls: str = '',
    ) -> None:
        self.file = file
        self.model_examples = model_examples
        self.model_examples_params = model_examples_params
        self.passport_model_data: dict[str, dict[str, Any]] = {}
        self.models_filename = models_filename
        self.parent_dict_cls = f'({parent_dict_cls})' if parent_dict_cls else ''

    def _construct_passports_model_cls(
        self, passports_modelname: str, dict_internal_pk_field: str
    ) -> str:
        """"""
        log.debug('Сборка класса модели паспортов.')

        model_cls_name = passports_modelname.rsplit('.', 1)[-1]

        ruff_noqa_line = '# ruff: noqa'

        cur_time = datetime.now(tz=timezone.utc).strftime('%Y-%m-%d %H:%M:%S')  # noqa: UP017
        time_line = f'# File creation time: {cur_time}\n\n'

        declaring_cls_line = f'class {model_cls_name}(models.Model):'

        doc_line = '    """Паспорты справочников НСИ."""\n'

        cls_lines = [ruff_noqa_line, time_line, declaring_cls_line, doc_line]

        for field_name, field_params in self.passport_model_data.items():
            params = []
            max_length = field_params.get('field_length')
            if max_length is not None:
                params.append(f'max_length={max_length}')

            if issubclass(field_params['field_type'], (dict, list)):
                params.append(f'default={field_params["field_type"].__name__}')

            if (
                self.model_examples_params is not None
                and 'all_fields_not_required' in self.model_examples_params
            ):
                if issubclass(field_params['field_type'], bool):
                    params.append('default=False')
                elif issubclass(field_params['field_type'], str):
                    params.append('blank=True')
                elif issubclass(field_params['field_type'], int):
                    params.append('null=True')

            field_cls_name = self.py_type_to_field_cls_name[field_params['field_type']]

            concat_params = ', '.join(params)
            model_field = f'    {field_name} = models.{field_cls_name}({concat_params})'
            cls_lines.append(model_field)

        declaring_meta_cls_line = '\n    class Meta:\n'
        meta_cls_fields = []
        ordering = f"        ordering = ('{dict_internal_pk_field}',)"
        meta_cls_fields.append(ordering)
        meta_cls_fields = '\n'.join(meta_cls_fields)
        meta_cls = f'{declaring_meta_cls_line}{meta_cls_fields}'
        cls_lines.append(meta_cls)

        return '\n'.join(cls_lines)

    def _construct_dict_model_cls(self, dict_state: DictState) -> str:
        """"""
        log.debug('Сборка класса модели справочника для: %s', dict_state.dict_filename)

        model_cls_name = dict_state.model.rsplit('.', 1)[-1]

        declaring_cls_line = f'class {model_cls_name}{self.parent_dict_cls}:'

        doc_line = f'    """{dict_state.passport_name}"""\n'

        cls_lines = [declaring_cls_line, doc_line]

        for field_name, field_params in dict_state.dict_model_data.items():
            params = []

            alias = field_params.get('alias')
            if issubclass(field_params['field_type'], (dict, list)):
                if alias:
                    params.append(f'verbose_name={alias!r}')
                params.append(f'default={field_params["field_type"].__name__}')
            else:
                if alias:
                    params.append(f'{alias!r}')

            max_length = field_params.get('field_length')
            if max_length is not None:
                params.append(f'max_length={max_length}')

            if (
                self.model_examples_params is not None
                and 'all_fields_not_required' in self.model_examples_params
            ):
                if issubclass(field_params['field_type'], bool):
                    params.append('default=False')
                elif issubclass(field_params['field_type'], str):
                    params.append('blank=True')
                elif issubclass(field_params['field_type'], int):
                    params.append('null=True')

            not_required_params = ('default=False', 'blank=True', 'null=True')
            if (
                not field_params.get('isrequired')
                and not issubclass(field_params['field_type'], str)
                and not any(p in params for p in not_required_params)
            ):
                params.append('null=True')

            field_cls_name = self.py_type_to_field_cls_name[field_params['field_type']]

            concat_params = ', '.join(params)
            model_field = f'    {field_name} = models.{field_cls_name}({concat_params})'
            cls_lines.append(model_field)

        for pk_name in self.DICT_PK_NAMES:
            if pk_name in dict_state.dict_model_data:
                nsi_pk_name = pk_name
                break
        else:
            raise NsiPkNotFoundError(
                f'Уникальный идентификатор не найден по ключам: {self.DICT_PK_NAMES!r} '
                f'в ключах справочника: {list(dict_state.dict_model_data.keys())}'
            )

        declaring_meta_cls_line = '\n    class Meta:\n'
        meta_cls_fields = []
        ordering = f"        ordering = ('{nsi_pk_name}',)"
        meta_cls_fields.append(ordering)
        meta_cls_fields = '\n'.join(meta_cls_fields)
        meta_cls = f'{declaring_meta_cls_line}{meta_cls_fields}'
        cls_lines.append(meta_cls)

        return '\n'.join(cls_lines)

    def upd_dict_model_data_with_passport_fields(
        self, dict_state: DictState, remote_passport: dict
    ) -> None:
        """"""
        if not self.model_examples:
            return
        log.debug('Обновление dict_model_data')

        for field_data in remote_passport['fields']:
            key: str = field_data['field']
            if key.lower() == 'extid':
                for pk_name in self.DICT_PK_NAMES:
                    if pk_name in dict_state.dict_model_data:
                        key = pk_name
                        break
                else:
                    raise NsiPkNotFoundError(
                        f'Уникальный идентификатор не найден по ключам: {self.DICT_PK_NAMES!r} '
                        f'в ключах справочника: {list(dict_state.dict_model_data.keys())}'
                    )

            for k in (
                'isrequired',
                'alias',
                # 'dataType',
            ):
                field_val = field_data[k]
                try:
                    dict_state.dict_model_data[key][k] = field_val
                except KeyError as exc:
                    log.warning(
                        'Ключ: %s есть в fields паспорта справочника, '
                        'но отсутствует в реальном файле справочника: %s '
                        'по этому, такое поле не будет добавлено в пример класса модели.',
                        exc,
                        dict_state.dict_filename,
                    )

        dict_state.dict_model_data = {
            k.replace('-', '_'): v for k, v in dict_state.dict_model_data.items()
        }

    def upd_dict_model_data_with_dict_fields(
        self, dict_state: DictState, dict_data: dict[str, Any]
    ) -> None:
        """"""
        if not self.model_examples:
            return

        for field_name, field_value in dict_data.items():
            #
            if field_name not in dict_state.dict_model_data:
                dict_state.dict_model_data[field_name] = {}

            type_field_value = type(field_value)
            if 'field_type' not in dict_state.dict_model_data[field_name]:
                dict_state.dict_model_data[field_name]['field_type'] = type_field_value

            if issubclass(type_field_value, str):
                len_field_value = len(field_value)

                if dict_state.dict_model_data[field_name].get('field_length', -1) < len_field_value:
                    dict_state.dict_model_data[field_name]['field_length'] = max(len_field_value, 1)

    def upd_passport_model_data_with_passport_fields(self, local_passports: list[dict]) -> None:
        """"""
        if not self.model_examples:
            return
        log.debug('Обновление passport_model_data')

        for passport in local_passports:
            #
            for field_name, field_value in passport['fields'].items():
                #
                if field_name not in self.passport_model_data:
                    self.passport_model_data[field_name] = {}

                type_field_value = type(field_value)
                if 'field_type' not in self.passport_model_data[field_name]:
                    self.passport_model_data[field_name]['field_type'] = type_field_value

                if issubclass(type_field_value, str):
                    len_field_value = len(field_value)

                    if self.passport_model_data[field_name].get('field_length', -1) < len_field_value:
                        self.passport_model_data[field_name]['field_length'] = max(len_field_value, 1)

        self.passport_model_data = {k.replace('-', '_'): v for k, v in self.passport_model_data.items()}

    def show_passport_model(
        self, passports_modelname: str, dict_internal_pk_field: str, passport_changed: bool
    ) -> None:
        """"""
        if not passport_changed:
            return

        if self.model_examples == 'stdout':
            model_cls = self._construct_passports_model_cls(passports_modelname, dict_internal_pk_field)
            print(model_cls)
        elif self.model_examples == 'file':
            model_cls = self._construct_passports_model_cls(passports_modelname, dict_internal_pk_field)
            self.file.remove_file(self.models_filename)
            self.file.write_text(self.models_filename, f'{model_cls}\n\n\n')

    def show_dict_model(self, dict_state: DictState) -> None:
        """"""
        if not dict_state.dict_changed:
            return

        if self.model_examples == 'stdout':
            model_cls = self._construct_dict_model_cls(dict_state)
            print(model_cls)
        elif self.model_examples == 'file':
            model_cls = self._construct_dict_model_cls(dict_state)
            self.file.write_text(self.models_filename, f'{model_cls}\n\n\n')



import abc
from collections.abc import Callable

try:
    from pypika import PostgreSQLQuery, Table
except ImportError:
    pass

from .dtos import DictState
from .logger import log


class Builder(abc.ABC):
    """"""

    @abc.abstractmethod
    def build_sql_copy_update_insert(
        self, dict_state: DictState, remote_dicts: list[dict], associate_map: dict[str, Callable]
    ) -> str:
        raise NotImplementedError


class SqlBuilder(Builder):
    """"""

    def __init__(
        self,
        dict_internal_pk_field: str,
        passports_rel_field: str,
    ) -> None:
        self.dict_internal_pk_field = dict_internal_pk_field
        # Название поля модели паспортов связанной со справочниками
        self.passports_rel_field = passports_rel_field
        self.sql_update_condition_fields = []

    def build_sql_copy_update_insert(
        self, dict_state: DictState, remote_dicts: list[dict], associate_map: dict[str, Callable]
    ) -> str:
        r"""
        Генерирует SQL формата::

            BEGIN;
            CREATE TEMP TABLE tmp_<таблица> AS TABLE <таблица> WITH NO DATA;
            ALTER TABLE <таблица> DISABLE TRIGGER ALL;
            COPY tmp_<таблица> ("<self.dict_internal_pk_field>", "field2", "field3") FROM STDIN WITH (FORMAT text);
            b962e385-610b-4d0a-806f-03311477a8f2	7026020	1.2.643.5.1.13.13.12.1.30.72291
            585193d7-d60a-4c5a-aa3e-d5cfd0f4bd77	7016244	1.2.643.5.1.13.13.12.4.16.63211
            \.

            UPDATE <таблица> AS t SET field2 = tmp.field2, field3 = tmp.field3 FROM tmp_<таблица> AS tmp WHERE t.<self.dict_internal_pk_field> = tmp.<self.dict_internal_pk_field>;
            INSERT INTO <таблица> SELECT * FROM tmp_<таблица> ON CONFLICT (<self.dict_internal_pk_field>) DO NOTHING;
            ALTER TABLE <таблица> ENABLE TRIGGER ALL;
            REINDEX TABLE <таблица>;
            COMMIT;

        https://www.postgresql.org/docs/current/sql-copy.html
        """  # noqa: E501
        log.debug('')

        sql_copy = self.build_sql_copy(dict_state, remote_dicts, associate_map)

        dict_model = dict_state.model.lower().replace('.', '_')

        fields_for_update = ', '.join(f'{f} = tmp.{f}' for f in associate_map)

        rows = []

        # Оборачиваем все запросы в одну транзакцию для ускорения
        begin_transaction = 'BEGIN;'
        # Создаем временную таблицу для обновления реальной таблицы
        create_temp_table = f'CREATE TEMP TABLE tmp_{dict_model} AS TABLE {dict_model} WITH NO DATA;'
        # Отключаем триггеры для ускорения вставки
        disable_trigger = f'ALTER TABLE {dict_model} DISABLE TRIGGER ALL;'

        # Позволяет определить в каком случае необходимо обновлять записи,
        # а именно когда значения в таблицах будут отличаться.
        if self.sql_update_condition_fields:
            conditions = ' OR '.join(
                f't.{f} IS DISTINCT FROM tmp.{f}' for f in self.sql_update_condition_fields
            )
            sql_update_condition = f' AND ({conditions})'
        else:
            sql_update_condition = ''

        # Обновляем записи
        update = (
            f'UPDATE {dict_model} AS t SET {fields_for_update} FROM tmp_{dict_model} AS tmp '
            f'WHERE t.{self.dict_internal_pk_field} = tmp.{self.dict_internal_pk_field}{sql_update_condition};'  # noqa: E501
        )
        # Добавляем новые записи если они появились
        insert = (
            f'INSERT INTO {dict_model} SELECT * FROM tmp_{dict_model} '
            f'ON CONFLICT ({self.dict_internal_pk_field}) DO NOTHING;'
        )
        # Включаем триггеры обратно
        enable_trigger = f'ALTER TABLE {dict_model} ENABLE TRIGGER ALL;'
        # Перестраиваем индексы
        reindex = f'REINDEX TABLE {dict_model};'
        # Завершаем транзакцию
        commit_transaction = 'COMMIT;'

        rows.append(begin_transaction)
        rows.append(create_temp_table)
        rows.append(disable_trigger)
        rows.append(sql_copy)
        rows.append(update)
        rows.append(insert)
        rows.append(enable_trigger)
        rows.append(reindex)
        rows.append(commit_transaction)

        return '\n'.join(rows)

    def build_sql_copy_merge_matched(self):
        r"""
        Генерирует SQL формата::

            BEGIN;
            CREATE TEMP TABLE tmp_<таблица> AS TABLE <таблица> WITH NO DATA;
            ALTER TABLE <таблица> DISABLE TRIGGER ALL;
            COPY tmp_<таблица> ("<self.dict_internal_pk_field>", "field2", "field3") FROM STDIN WITH (FORMAT text);
            b962e385-610b-4d0a-806f-03311477a8f2	7026020	1.2.643.5.1.13.13.12.1.30.72291
            585193d7-d60a-4c5a-aa3e-d5cfd0f4bd77	7016244	1.2.643.5.1.13.13.12.4.16.63211
            \.

            MERGE INTO <таблица> AS t USING tmp_<таблица> AS tmp ON t.<self.dict_internal_pk_field> = tmp.<self.dict_internal_pk_field>
            WHEN MATCHED THEN UPDATE SET field2 = tmp.field2, field3 = tmp.field3
            WHEN NOT MATCHED THEN INSERT (<self.dict_internal_pk_field>, field2, field3) VALUES (tmp.<self.dict_internal_pk_field>, tmp.field2, tmp.field3);
            ALTER TABLE <таблица> ENABLE TRIGGER ALL;
            REINDEX TABLE <таблица>;
            COMMIT;
        """  # noqa: E501

    def build_sql_copy(
        self, dict_state: DictState, remote_dicts: list[dict], associate_map: dict[str, Callable]
    ) -> str:
        r"""
        Генерирует SQL формата::

            COPY public.<таблица> ("<self.dict_internal_pk_field>", "field2", "field3") FROM STDIN WITH (FORMAT text);
            b962e385-610b-4d0a-806f-03311477a8f2	7026020	1.2.643.5.1.13.13.12.1.30.72291
            585193d7-d60a-4c5a-aa3e-d5cfd0f4bd77	7016244	1.2.643.5.1.13.13.12.4.16.63211
            \.

        https://www.postgresql.org/docs/current/sql-copy.html
        """  # noqa: E501
        log.debug('')

        self._changing_defaults_for_copy(associate_map)

        self._data_preparation(remote_dicts, associate_map)

        dict_model = dict_state.model.lower().replace('.', '_')
        # public_dict_model = f'public.{dict_model}'

        # NOTE(Ars): Для SQL фикстуры не получиться использовать pk как алиас для dict_internal_pk_field,
        # поля должны совпадать по имени с тем что действительно записано в БД.
        fields = (self.dict_internal_pk_field, *associate_map.keys())

        # NOTE(Ars): Postgres приводит все названия полей к нижнему регистру
        # если не поместить их в двойные кавычки.
        solid_fields = ', '.join(f'"{f}"' for f in fields)

        start_of_data_marker = ';'
        end_of_data_marker = '\\.\n'

        sql_query = (
            f'COPY {dict_model} ({solid_fields}) FROM STDIN WITH (FORMAT text){start_of_data_marker}'
        )

        rows = [sql_query]
        for remote_dict in remote_dicts:
            values = '\t'.join(str(v) for v in remote_dict['fields'].values())
            row = f'{remote_dict["pk"]}\t{values}'
            rows.append(row)

        rows.append(end_of_data_marker)

        return '\n'.join(rows)

    def build_sql_insert_or_update(
        self, dict_state: DictState, remote_dicts: list[dict], associate_map: dict[str, Callable]
    ) -> str:
        """
        Генерирует SQL формата::

            INSERT INTO "<таблица>" ("<self.dict_internal_pk_field>","field2","field3") VALUES ('7bf3449a-2ae8-49cc-bdc6-c380f7aa1918',362,'239'),('df349786-c9f0-485f-9326-2946b0f21424',364,'239') ON CONFLICT ("<self.dict_internal_pk_field>") DO UPDATE SET "field2"=EXCLUDED."field2","field3"=EXCLUDED."field3";
        """  # noqa: E501
        log.debug('')

        self._data_preparation(remote_dicts, associate_map)

        dict_model = dict_state.model.lower().replace('.', '_')

        # NOTE(Ars): Для SQL фикстуры не получиться использовать pk как алиас для dict_internal_pk_field,
        # поля должны совпадать по имени с тем что действительно записано в БД.
        fields = (self.dict_internal_pk_field, *associate_map.keys())

        table = Table(dict_model)
        values = ((remote_dict['pk'], *remote_dict['fields'].values()) for remote_dict in remote_dicts)
        query = (
            PostgreSQLQuery.into(table)
            .columns(*fields)
            .insert(*values)
            .on_conflict(getattr(table, self.dict_internal_pk_field))
        )
        for field_name in associate_map:
            query = query.do_update(field_name, None)  # None это EXCLUDED."field_name"
        return f'{query.get_sql()};'

    def _data_preparation(self, remote_dicts: list[dict], associate_map: dict[str, Callable]) -> None:
        """"""
        # NOTE(Ars): Возможно для стабильности работы придется сортировать ключи
        # по своему усмотрению, чтобы nsi не начудил

        # NOTE(Ars): Для SQL фикстуры не получиться использовать passports_rel_field
        # как алиас для passports_rel_field + _id, поля должны
        # совпадать по имени с тем что действительно записано в БД.
        associate_map[f'{self.passports_rel_field}_id'] = associate_map.pop(self.passports_rel_field)
        for remote_dict in remote_dicts:
            remote_dict['fields'] = {
                f'{self.passports_rel_field}_id': remote_dict['fields'].pop(self.passports_rel_field),
                # Для того чтобы passports_rel_field + _id переместился в начало словаря
                **remote_dict['fields'],
            }

        # Добавления недостающих полей
        for remote_dict in remote_dicts:
            dict_fields: dict = remote_dict['fields']
            updated_fields = {}

            for field_name, field_type in associate_map.items():
                if field_name in dict_fields:
                    updated_fields[field_name] = dict_fields[field_name]
                else:
                    # Добавляем значение по умолчанию для недостающего поля
                    updated_fields[field_name] = field_type()

            # Обновляем поля в правильном порядке
            remote_dict['fields'] = updated_fields

    def _changing_defaults_for_copy(self, associate_map: dict[str, Callable]) -> None:
        """"""
        for field_name, field_type in associate_map.items():
            # NOTE(Ars): NULL в SQL файле для COPY по умолчанию записываеться как '\N'
            if field_type() is None:
                associate_map[field_name] = lambda: r'\N'
            # NOTE(Ars): Пустая строка "" в файле представлена как два знака табуляции: \t\t
            # Пустая строка это отсутствие знака между
            # двумя разделительными табуляциями \t""\t == \t\t\t\t
            elif isinstance(field_type, str):
                associate_map[field_name] = lambda: r'\t\t'



import pytest
from click.testing import CliRunner
from src.loadnsi.loadnsi import loadnsi


@pytest.mark.skip(reason='TODO')
def test_loadnsi():
    runner = CliRunner()
    result = runner.invoke(
        loadnsi,
        # loadnsi --log_level=DEBUG --compress_files=gzip --use_official_api --model_examples=file --model_examples_params=all_fields_not_required --forced_update organization_nsi department_nsi  # noqa: E501
        # [
        #     '--log_level', 'DEBUG',
        #     '--compress_files', 'gzip',
        #     '--use_official_api',  # Флаговая опция (без значения)
        #     '--model_examples', 'file',
        #     '--model_examples_params', 'all_fields_not_required',
        #     '--forced_update', 'organization_nsi', 'department_nsi',  # Последовательные значения
        # ],
    )
    print(result)
    # print(result.exit_code)
    # print(result.exception)
    # print(result.output)



import shutil
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

import httpx
from rich.console import Console, Group
from rich.live import Live
from rich.logging import RichHandler
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)
from rich.text import Text

from .dtos import DictState

rich_console = Console()


def update_console_size() -> None:
    """
    Обновляет размеры консоли, чтобы убрать дублирование элементов
    при изменении размера терминала.
    """
    _width, _height = shutil.get_terminal_size()
    rich_console.size = (_width - 1, _height - 4)


log_handler = RichHandler(
    # level=options['log_level'],
    console=rich_console,
    show_time=False,
    # show_level=False,
    show_path=False,
    rich_tracebacks=True,
)

download_progress_bar = Progress(
    TextColumn('[cyan][progress.description]{task.description}'),
    BarColumn(bar_width=None),
    TextColumn('[progress.percentage]{task.percentage:>3.0f}%'),
    DownloadColumn(),
    TransferSpeedColumn(),
    TimeRemainingColumn(),
)
timeout_progress_bar = Progress(
    SpinnerColumn(),
    TextColumn('[cyan][progress.description]{task.description}'),
)
retry_wait_progress_bar = Progress(
    SpinnerColumn(),
    TextColumn('[cyan][progress.description]{task.description}'),
)

rich_panel1 = Panel(
    retry_wait_progress_bar, title='Сон перед повторной попыткой', border_style='yellow', expand=True
)
rich_panel2 = Panel(
    timeout_progress_bar, title='Ожидание ответа от сервера', border_style='blue', expand=True
)

download_text = Text()
download_group = Group(download_text, download_progress_bar)
rich_panel3 = Panel(download_group, title='Загрузка справочников', border_style='green', expand=True)

rich_group = Group(rich_panel1, rich_panel2, rich_panel3)

rich_live = Live(rich_group, console=rich_console)


@asynccontextmanager
async def rich_track_download(dict_state: DictState, response: httpx.Response):
    """
    Контекстный менеджер для отслеживания прогресса загрузки с помощью rich.
    """
    description = f'Загрузка справочника: {dict_state.dict_filename}'

    task_id = download_progress_bar.add_task(
        description=f'[white]№{0:2}:[/white] {description}',
        total=int(response.headers.get('Content-Length', 0)),
    )

    if not download_text._text[0]:
        download_text._text[0] = f'Всего справочников: {dict_state.total_dicts}шт.'

    original_aiter_bytes = response.aiter_bytes

    async def aiter_bytes_with_progress(chunk_size: int | None = None) -> AsyncIterator[bytes]:
        async for chunk in original_aiter_bytes(chunk_size):
            yield chunk
            download_progress_bar.update(
                task_id,
                description=(f'[white]№{task_id:2}:[/white] {description}'),
                completed=response.num_bytes_downloaded,
                # advance=len(chunk),
            )
            update_console_size()  # XXX(Ars): Надеюсь это не замедлит загрузку..

    response.aiter_bytes = aiter_bytes_with_progress
    try:
        yield task_id
    finally:
        response.aiter_bytes = original_aiter_bytes


@asynccontextmanager
async def rich_track_timeout(description: str, client: httpx.AsyncClient):
    """"""
    task_id = timeout_progress_bar.add_task(
        description=f'{description} до {client.timeout.connect}сек.',
    )
    try:
        yield task_id
    finally:
        timeout_progress_bar.remove_task(task_id)


@asynccontextmanager
async def rich_track_retry_wait(description: str, total: float):
    """"""
    task_id = retry_wait_progress_bar.add_task(
        description=f'{description} {total:.1f}сек.',
    )
    try:
        yield task_id
    finally:
        retry_wait_progress_bar.remove_task(task_id)



import abc
import io
import json
import logging

import httpx
from bs4 import BeautifulSoup
from tenacity import (
    RetryCallState,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,  # noqa: F401
    wait_random,
)

from .console_view import (
    rich_track_download,
    rich_track_retry_wait,
    rich_track_timeout,
)
from .dtos import DictState
from .exceptions import (
    NsiBadResultError,
    NsiFileNotFoundError,
    NsiScriptTagNotFoundError,
)
from .logger import log

MAX_ATTEMPT_NUMBER = 4


class SafeDictFormatMap(dict):
    """Всегда возвращает исходный ключ форматирования `...{key}...` для пропущенных ключей."""

    def __missing__(self, key) -> str:
        return f'{{{key}}}'


def parse_build_id_from_html(text: str) -> str:
    soup = BeautifulSoup(text, 'html.parser')
    # Ищем элемент: <script id="__NEXT_DATA__" type="application/json">
    script_tag = soup.find('script', id='__NEXT_DATA__', type='application/json')
    if not script_tag:
        raise NsiScriptTagNotFoundError(
            'Тег: <script id="__NEXT_DATA__" type="application/json"> не найден.'
        )
    # Извлекаем текстовое содержимое тега
    tag_data: dict = json.loads(script_tag.string)
    return tag_data.get('buildId')


async def handle_retry_error_callback(retry_state: RetryCallState) -> None:
    """"""
    log.debug(f'{retry_state}; args: {retry_state.args[1:]};')
    # Завершение контекста если retry исчерпал все попытки (handle_before не срабатывает)
    await retry_state.rich_track_retry_wait_ctx.__aexit__(None, None, None)
    # Проброс последнего исключения наверх, чтобы его поймал gather и отменил задачу
    raise retry_state.outcome.exception()


async def handle_before_sleep(retry_state: RetryCallState) -> None:
    """"""
    log.debug(f'{retry_state}; args: {retry_state.args[1:]};')


async def handle_before(retry_state: RetryCallState) -> None:
    """"""
    # if retry_state.idle_for == 0.0:
    # if retry_state.upcoming_sleep == 0.0:
    if retry_state.attempt_number == 1:
        # Пропускаем первый вызов пустышку, хз зачем он сделан.
        return
    await retry_state.rich_track_retry_wait_ctx.__aexit__(None, None, None)


async def handle_after(retry_state: RetryCallState) -> None:
    """"""
    dict_state: DictState = retry_state.args[1]
    description = (
        f'Сон №{retry_state.attempt_number}/{MAX_ATTEMPT_NUMBER} '
        f'перед попыткой повторного запроса: {dict_state.dict_filename}'
    )
    retry_state.rich_track_retry_wait_ctx = rich_track_retry_wait(
        description,
        retry_state.upcoming_sleep,  # retry_state.idle_for
    )
    await retry_state.rich_track_retry_wait_ctx.__aenter__()


# NOTE(Ars): На каждый метод должен использоваться свой экземпляр декоратора @retry
retry_params = dict(
    # Повтор только при конкретных исключениях
    retry=retry_if_exception_type((httpx.HTTPError, NsiBadResultError)),
    stop=stop_after_attempt(MAX_ATTEMPT_NUMBER),  # Делаем 4 попытки
    wait=wait_random(min=3, max=6),  # Ждем от 3 до 6 секунд перед следующей попыткой
    # wait=tenacity.wait_fixed(5),  # Ждем ровно N секунд перед следующей попыткой
    # wait=wait_exponential(multiplier=0.2, max=10),  # Ждём 0.2 -> 0.4 -> 0.8 -> 1.6 -> 3.2 и т.д.
    retry_error_callback=handle_retry_error_callback,  # Вызывается перед wait последней попытки
    before_sleep=handle_before_sleep,  # Вызывается перед wait каждой попытки кроме последней
    before=handle_before,
    after=handle_after,
)


class NsiWebCrawler(abc.ABC):
    """Базовый класс для действий связанных с загрузкой данных из внешней системы по HTTP."""

    def __init__(
        self,
        nsi_base_url: str,
        nsi_versions_path: str,
        nsi_passport_path: str,
        nsi_files_path: str,
    ) -> None:
        self.nsi_versions_path = nsi_versions_path
        self.nsi_passport_path = nsi_passport_path
        self.nsi_files_path = nsi_files_path

        self.data_client_kw = {
            'params': {},
            'headers': {
                'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 '
                '(KHTML, like Gecko) Chrome/106.0.0.0 Safari/537.36',
            },
            'verify': False,
            'timeout': httpx.Timeout(35),
            'follow_redirects': True,
            'base_url': nsi_base_url,
        }

        self.file_client_kw = self.data_client_kw.copy()

    @abc.abstractmethod
    async def get_remote_passport(self, dict_state: DictState, get_kwargs: dict | None = None) -> dict:
        """Возвращает паспорт справочника."""
        log.debug('Запрос паспорта для: %s', dict_state.dict_filename)
        async with httpx.AsyncClient(**self.data_client_kw) as client:
            #
            description = f'Ожидание ответа на получение паспорта: {dict_state.dict_filename}'
            async with rich_track_timeout(description, client):
                response: httpx.Response = await client.get(**get_kwargs)

            log.debug('status_code: %s', response.status_code)
            response.raise_for_status()
            return response.json()

    @abc.abstractmethod
    async def get_dict_versions(self, dict_state: DictState) -> dict:
        """Возвращает информацию о версиях справочника."""
        log.debug('Запрос для: %s', dict_state.dict_filename)
        async with httpx.AsyncClient(**self.data_client_kw) as client:
            #
            description = f'Ожидание ответа на получение версий: {dict_state.dict_filename}'
            async with rich_track_timeout(description, client):
                response = await client.get(
                    self.nsi_versions_path, params={'identifier': dict_state.oid}
                )

            log.debug('status_code: %s', response.status_code)
            response.raise_for_status()
            return response.json()

    @retry(**retry_params)
    async def get_zip_filename(self, dict_state: DictState, version: str, format: str = 'JSON') -> str:  # noqa: A002
        """Формирует название zip архива."""
        log.debug('Запрос на получение названия архива для: %s', dict_state.dict_filename)
        async with httpx.AsyncClient(**self.file_client_kw) as client:
            #
            description = f'Ожидание ответа на получение названия архива: {dict_state.dict_filename}'
            async with rich_track_timeout(description, client):
                response = await client.get(
                    self.nsi_files_path,
                    params={'identifier': dict_state.oid, 'version': version, 'format': format},
                )

            log.debug('status_code: %s', response.status_code)
            response.raise_for_status()

            data = response.json()
            if not data:
                raise NsiFileNotFoundError(
                    f'Файл с версией: {version} не найден, '
                    'возможно он ещё не создан в системе NSI по крону.'
                )
            return data[0]

    @retry(**retry_params)
    async def download_zip(self, dict_state: DictState, zip_filename: str) -> io.BytesIO:
        """Скачивает zip архив в потоковом режиме с отображением прогресса."""
        log.debug('Скачивание для: %s Архива: %s', dict_state.dict_filename, zip_filename)
        zip_buffer = io.BytesIO()

        async with httpx.AsyncClient(**self.file_client_kw) as client:
            request = client.build_request('GET', f'{self.nsi_files_path}/{zip_filename}')

            description = f'Ожидание ответа на получение файла архива: {dict_state.dict_filename}'
            async with rich_track_timeout(description, client):
                response = await client.send(request, stream=True)

            log.debug('status_code: %s', response.status_code)
            response.raise_for_status()

            async with rich_track_download(dict_state, response):
                async for chunk in response.aiter_bytes():
                    zip_buffer.write(chunk)

            await response.aclose()

        return zip_buffer


class OfficialApiNsiWebCrawler(NsiWebCrawler):
    """
    Класс для действий связанных с загрузкой данных из внешней системы по HTTP.
    Работающий с официальным апи НСИ который использует аутентификацию по токену.

    https://nsi.rosminzdrav.ru/port/swagger-ui.html
    """

    def __init__(
        self,
        nsi_base_url: str,
        nsi_versions_path: str,
        nsi_passport_path: str,
        nsi_files_path: str,
        nsi_api_user_key: str | None = None,
    ) -> None:
        super().__init__(nsi_base_url, nsi_versions_path, nsi_passport_path, nsi_files_path)

        if nsi_api_user_key:
            self.data_client_kw['params'].update({'userKey': nsi_api_user_key})

    @retry(**retry_params)
    async def get_remote_passport(self, dict_state: DictState, get_kwargs: dict | None = None) -> dict:
        get_kwargs = {'url': self.nsi_passport_path, 'params': {'identifier': dict_state.oid}}
        data = await super().get_remote_passport(dict_state, get_kwargs)
        self._raise_for_result(data)
        return data

    @retry(**retry_params)
    async def get_dict_versions(self, dict_state: DictState) -> dict:
        data = await super().get_dict_versions(dict_state)
        self._raise_for_result(data)
        return data

    @staticmethod
    def _raise_for_result(data: dict) -> None:
        # {'result': 'ERROR', 'resultText': 'Запрашиваемая версия не существует', 'resultCode': '03x0006'}  # noqa: E501
        if data['result'] != 'OK':
            raise NsiBadResultError(f'{data!r}')


class PirateApiNsiWebCrawler(NsiWebCrawler):
    """
    Класс для действий связанных с загрузкой данных из внешней системы по HTTP.
    Работающий с пиратским апи НСИ для которого не требуеться аутентификация.
    """

    def __init__(
        self,
        nsi_base_url: str,
        nsi_versions_path: str,
        nsi_passport_path: str,
        nsi_files_path: str,
    ) -> None:
        super().__init__(nsi_base_url, nsi_versions_path, nsi_passport_path, nsi_files_path)
        try:
            resp_text = self._get_html_with_build_id()
        except Exception as exc:
            if log.level == logging.DEBUG:
                raise exc
            raise type(exc)('Не удалось получить buildId, попробуйте повторить попытку.')  # noqa: B904

        build_id = parse_build_id_from_html(resp_text)

        self.nsi_passport_path = self.nsi_passport_path.format_map(SafeDictFormatMap(buildId=build_id))

    @retry(
        retry=retry_if_exception_type((httpx.HTTPError,)),
        stop=stop_after_attempt(MAX_ATTEMPT_NUMBER),
        wait=wait_random(min=3, max=6),
    )
    def _get_html_with_build_id(self) -> str:
        """Возвращает buildId для построения url path к апи получения passport."""
        log.debug('Запрос для получения buildId')
        data_client_kw = dict(self.data_client_kw, timeout=httpx.Timeout(15))
        with httpx.Client(**data_client_kw) as client:
            response = client.get('')
            log.debug('status_code: %s', response.status_code)
            response.raise_for_status()
        return response.text

    @retry(**retry_params)
    async def get_remote_passport(self, dict_state: DictState, get_kwargs: dict | None = None) -> dict:
        get_kwargs = {'url': self.nsi_passport_path.format(identifier=dict_state.oid, version='latest')}
        data = await super().get_remote_passport(dict_state, get_kwargs)
        return data['pageProps']['dict']

    @retry(**retry_params)
    async def get_dict_versions(self, dict_state: DictState) -> dict:
        return await super().get_dict_versions(dict_state)



NSI_FIXTURES_FOLDER = 'fixtures/nsi'
NSI_PASSPORTS = {
    'file': 'dictionary_nsi.json',
    'model': 'nsi.DictionaryNsi',
    # 'include': ('oid', 'version', 'fullName'),
    # 'exclude': ('fullName', 'shortName'),
}
DICT_INTERNAL_PK = 'custom_id'
PASSPORTS_REL = 'nsi_dictionary'
PARENT_DICT_CLS = 'BaseNsi'
NSI_DICTIONARIES = {
    'organization_nsi.json': {
        'model': 'nsi.OrganizationNsi',
        'oid': '1.2.643.5.1.13.13.11.1461',
        # 'filter': lambda r: r.get('regionName') == 'Астраханская область',
        'create_sql': True,
    },
    'department_nsi.json': {
        'model': 'nsi.DepartmentNsi',
        'oid': '1.2.643.5.1.13.13.99.2.114',
        'create_sql': True,
    },
    'position_nsi.json': {
        'model': 'nsi.PositionNsi',
        'oid': '1.2.643.5.1.13.13.99.2.181',
    },
    'ensurance.json': {
        'model': 'nsi.EnsuranceNsi',
        'oid': '1.2.643.5.1.13.13.99.2.183',
    },
    'placement.json': {
        'model': 'nsi.PlacementNsi',
        'oid': '1.2.643.5.1.13.13.99.2.322',
    },
    'payment.json': {
        'model': 'nsi.PaymentNsi',
        'oid': '1.2.643.5.1.13.13.11.1039',
    },
    'subjects_rf.json': {
        'model': 'nsi.SubjectRF',
        'oid': '1.2.643.5.1.13.13.99.2.206',
    },
    'document_types_nsi.json': {
        'model': 'nsi.DocumentTypesNsi',
        'oid': '1.2.643.5.1.13.13.11.1522',
    },
    'medical_card_types_nsi.json': {
        'model': 'nsi.MedicalCardTypesNsi',
        'oid': '1.2.643.5.1.13.13.99.2.723',
    },
    'nosological_diagnosis_types_nsi.json': {
        'model': 'nsi.NosologicalDiagnosisTypesNsi',
        'oid': '1.2.643.5.1.13.13.11.1077',
    },
    'family_status_nsi.json': {
        'model': 'nsi.FamilyStatusNsi',
        'oid': '1.2.643.5.1.13.13.99.2.15',
    },
    'education_nsi.json': {
        'model': 'nsi.EducationNsi',
        'oid': '1.2.643.5.1.13.13.99.2.16',
    },
    'employment_nsi.json': {
        'model': 'nsi.EmploymentNsi',
        'oid': '1.2.643.5.1.13.13.11.1038',
    },
    'identity_doc_nsi.json': {
        'model': 'nsi.IdentityDocNsi',
        'oid': '1.2.643.5.1.13.13.99.2.48',
    },
    'family_relationship.json': {
        'model': 'nsi.FamilyRelationship',
        'oid': '1.2.643.5.1.13.13.99.2.14',
    },
    'legal_guardian_doc_dictionary.json': {
        'model': 'nsi.LegalGuardianDocNsi',
        'oid': '1.2.643.5.1.13.13.99.2.313',
        # 'filter': lambda r: 'Свидетельство' in r.get('name', ''),
        # 'include': ('ID', 'REPRESENTED', 'ATTRIBUTE_FACE'),
        # 'exclude': ('REPRESENTED', 'ATTRIBUTE_FACE'),
    },
    'address_type_nsi.json': {
        'model': 'nsi.AddressTypeNsi',
        'oid': '1.2.643.5.1.13.13.11.1504',
    },
    'oms_police_type_nsi.json': {
        'model': 'nsi.OmsPoliceTypeNsi',
        'oid': '1.2.643.5.1.13.13.11.1035',
    },
}
